/*
 * MacSamba
 * <http://ghostwood.org/software/macSamba.html>
 * Copyright 2006-2015 Nick Howes <nick@ghostwood.org>
 * tabs every 4 spaces
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <IOKit/usb/IOUSBLib.h>

#include "util.h"
#include "devices.h"
#include "usb.h"
#include "ops.h"

#define BUFFER_SIZE		0x00200000 //2MB

#define GREET_INFO "MacSamba 2.0 (for use with Atmel ARM Cortex-M7 devices) (c)2006-2015 Nick Howes\n"
#define USAGE_INFO "MacSamba [-k] [-s] [-o OFFSET] [-c COUNT] [-f FILE] [-l] [-a] [-d] [-e] [-p] [-w] [-m] [-r]\n"
#define OPT_INFO_1  "  -k         disables the Apple USB CDC kext (if active) before proceeding (requires superuser)\n"
#define OPT_INFO_2  "  -s         enables the Apple USB CDC kext after completion (requires superuser)\n"
#define OPT_INFO_3  "  -o OFFSET  sets OFFSET for reading/writing/programming (bytes)\n"
#define OPT_INFO_4  "  -c COUNT   sets COUNT for reading/writing/programming (bytes)\n"
#define OPT_INFO_5  "  -f FILE    reads raw binary data file into buffer, also sets count\n"
#define OPT_INFO_6  "  -l         reads from flash into buffer\n"
#define OPT_INFO_7  "  -a         reads from RAM into buffer\n"
#define OPT_INFO_8  "  -d         dumps the buffer\n"
#define OPT_INFO_9  "  -e         erases the flash memory\n"
#define OPT_INFO_10 "  -p         programs the buffer into flash memory\n"
#define OPT_INFO_11 "  -w         writes the buffer into RAM memory\n"
#define OPT_INFO_12 "  -m         re-maps (sets GPNVM 1) to boot from flash\n"
#define OPT_INFO_13 "  -r         resets the CPU after completing other tasks\n"

#define EEFC_FMR_ADDR				0x400E0C00	// Flash Mode Register - not used

#define EEFC_FCR_ADDR				0x400E0C04	// Flash Command Register
#define EEFC_FCR_CMD_ERASE_ALL		0x5A000005
#define EEFC_FCR_CMD_WRITE_PAGE		0x5A000001 // | (page << 8)
#define EEFC_FCR_CMD_SET_GPNVM		0x5A00000B // | (bit << 8)

#define RSTC_CR_ADDR				0x400E1800	// Reset Control Register
#define RSTC_CR_CMD_RST_ALL			0xA5000009  // processor + external reset


#define KEXT_LINE_LEN	120
#define KEXTNAME		"com.apple.driver.AppleUSBCDC"
#define KEXTPATH		"/System/Library/Extensions/IOUSBFamily.kext/Contents/PlugIns/AppleUSBCDC.kext"


//-----------------------------------------------------------------------------------------------------------------

int disableCDC();
int enableCDC();
int checkCDC();

void doCore(int offset, int count, int fd, int readFlash, int readRam, int dump, int erase, int program, int writeRam, int remap, int reset);

//-----------------------------------------------------------------------------------------------------------------

int main(int argc, char **argv) {
	int kextUnload = 0;
	int kextLoad = 0;
	int offset = 0;
	int count = 0;
	int fd = -1;
	int readFlash = 0;
	int readRam = 0;
	int dump = 0;
	int erase = 0;
	int program = 0;
	int writeRam = 0;
	int remap = 0;
	int reset = 0;
	
	int k;
	extern char *optarg;
	extern int optind;
	int ch;
	
	printf(GREET_INFO);
	
	//handle user input
	while ((ch = getopt(argc, argv, "kso:c:f:ladepwmrh")) != -1) {
		switch (ch) {
			case 'k':
				kextUnload = 1;
				break;
			case 's':
				kextLoad = 1;
				break;
			case 'o':
				offset = atoi(optarg);
				break;
			case 'c':
				count = atoi(optarg);
				break;
			case 'f':
				fd = open(optarg, O_RDONLY);
				if (fd < 0) failExit("cannot open specified file", 0);
				break;
			case 'l':
				readFlash = 1;
				break;
			case 'a':
				readRam = 1;
				break;
			case 'd':
				dump = 1;
				break;
			case 'e':
				erase = 1;
				break;
			case 'p':
				program = 1;
				break;
			case 'w':
				writeRam = 1;
				break;
			case 'm':
				remap = 1;
				break;
			case 'r':
				reset = 1;
				break;
			default:
				printf (USAGE_INFO);
				printf (OPT_INFO_1);
				printf (OPT_INFO_2);
				printf (OPT_INFO_3);
				printf (OPT_INFO_4);
				printf (OPT_INFO_5);
				printf (OPT_INFO_6);
				printf (OPT_INFO_7);
				printf (OPT_INFO_8);
				printf (OPT_INFO_9);
				printf (OPT_INFO_10);
				printf (OPT_INFO_11);
				printf (OPT_INFO_12);
				printf (OPT_INFO_13);
				exit(2);
		}
	}
	
	//handle kext unload
	if (kextUnload) {
		k = checkCDC();
		if (k == -1) {
			failExit("Unable to determine AppleUSBCDC kext state", -1);
		} else if (k == 1) {
			disableCDC();
			printf("AppleUSBCDC kext disabled\n");
		} else {
			printf("Note: command ignored: AppleUSBCDC kext was not active\n");		
		}
	}
	
	//core stuff
	doCore(offset, count, fd, readFlash, readRam, dump, erase, program, writeRam, remap, reset);
	
	//handle kext load
	if (kextLoad) {
		k = checkCDC();
		if (k == -1) {
			failExit("Unable to determine AppleUSBCDC kext state", -1);
		} else if (k == 0) {
			enableCDC();
			printf("AppleUSBCDC kext enabled\n");
		} else {
			printf("Note: command ignored: AppleUSBCDC kext was already active\n");		
		}
	}

	//clean up	
	if (fd >= 0) close(fd);
	exit(0);
	
}


//-----------------------------------------------------------------------------------------------------------------

// this function checks to see if the USB CDC kernel extension is active
// return 1 if it is; 0 if it is not, -1 otherwise
int checkCDC() {
	FILE *fp;
	char line[KEXT_LINE_LEN];
	int rslt = 0;

	fp = popen("kextstat -l -b com.apple.driver.AppleUSBCDC", "r");
	if (fp == NULL) {
		rslt = -1;
	} else {
		while (fgets(line, KEXT_LINE_LEN, fp) != NULL) {
			rslt = 1;
			break;
		}
		pclose(fp);
	}
	
	return rslt;
}


//-----------------------------------------------------------------------------------------------------------------

int disableCDC() {
	FILE *fp = popen("kextunload /System/Library/Extensions/IOUSBFamily.kext/Contents/PlugIns/AppleUSBCDC.kext", "r");
	return pclose(fp);
}


//-----------------------------------------------------------------------------------------------------------------

int enableCDC() {
	FILE *fp = popen("kextload /System/Library/Extensions/IOUSBFamily.kext/Contents/PlugIns/AppleUSBCDC.kext", "r");
	return pclose(fp);
}


//-----------------------------------------------------------------------------------------------------------------

void doCore(int offset, int count, int fd, int readFlash, int readRam, int dump, int erase, int program, int writeRam, int remap, int reset) {
	char buffer[BUFFER_SIZE];
	int i, j;
	usb_t usb;
	unsigned int data;
	int readLen;
	unsigned int addr, page;
	deviceInfo_t* dev = NULL;
	int dotCounter = 18;

	//setup connection and device
	if (readFlash || readRam || erase || program || writeRam || remap || reset) {
		usb = getUSB();
		dev = getDeviceInfo(usb);
		if (dev == NULL) failExit("Unknown CPU\n", 0);
		printf("Device is %s\n", dev->name);
		//readWord(usb, EEFC_FMR_ADDR, &data);
		//printf("Flash Mode Register is 0x%08X\n", data);
	}
	
	//handle file
	if (fd >= 0) {
		printf("Reading in file ...");
		fflush(stdout);
		i = 0;
		for (;;) {
			readLen = read(fd, buffer + i, sizeof(buffer));
			if (readLen == 0) break;
			if (readLen < 0) failExit("file read failed\n", 0);
			i += readLen;
		}
		printf(" OK\n");
		if (count != 0) printf("WARNING: resetting count to file length (was 0x%08X)\n", count);
		count = i;
	}
	
	//count and offset checks
	if (offset < 0) failExit("ERROR: offset is negative", -2);
	if (count >= BUFFER_SIZE) failExit("ERROR: count too large", -3);
	if ((offset % 4) != 0) printf("WARNING: offset is not word-aligned (0x%08X)\n", offset);
	if ((count % 4) != 0) printf("WARNING: count is not word-aligned (0x%08X)\n", count);
	if (offset != 0) printf("Note: offset is 0x%08X\n", offset);
	if (count != 0) printf("Note: count is 0x%08X\n", count);
	
	//handle read flash
	if (readFlash) {
		printf("Reading from flash ");
		fflush(stdout);
		dotCounter = 18;
		addr = FLASH_ADDR + offset;
		for (i=0 ; i<count ; i+=4) {
			readWord(usb, addr+i, &data);
			buffer[i]   =  data        & 0xff;
			buffer[i+1] = (data >>  8) & 0xff;
			buffer[i+2] = (data >> 16) & 0xff;
			buffer[i+3] = (data >> 24) & 0xff;
			if (i%512 == 0) {
				printf(".");
				if (++dotCounter > 80) {
					printf("\n");
					dotCounter = 0;
				}
				fflush(stdout);
			}
		}
		printf(" OK\n");
	}
	
	//handle read RAM
	if (readRam) {
		printf("Reading from RAM ");
		fflush(stdout);
		dotCounter = 16;
		addr = RAM_ADDR + offset;
		for (i=0 ; i<count ; i+=4) {
			readWord(usb, addr+i, &data);
			buffer[i]   =  data        & 0xff;
			buffer[i+1] = (data >>  8) & 0xff;
			buffer[i+2] = (data >> 16) & 0xff;
			buffer[i+3] = (data >> 24) & 0xff;
			if (i%512 == 0) {
				printf(".");
				if (++dotCounter > 80) {
					printf("\n");
					dotCounter = 0;
				}
				fflush(stdout);
			}
		}
		printf(" OK\n");
	}
	
	//handle dump
	if (dump) {
		for (i=0 ; i<count ; i+=4) {
			data  =  buffer[i]   & 0xff;
			data |= (buffer[i+1] & 0xff) <<  8;
			data |= (buffer[i+2] & 0xff) << 16;
			data |= (buffer[i+3] & 0xff) << 24;
			if (i%16 == 0) {
				printf("\n");
				printf("%08X : ", i + offset);
			}
			printf("%08X ", data);
		}
		printf("\n");
	}

	//handle erase
	if (erase) {
		printf("Erasing ...");
		fflush(stdout);
		writeWord(usb, EEFC_FCR_ADDR, EEFC_FCR_CMD_ERASE_ALL);
		waitFlashReady(usb);
		printf(" OK\n");
	}
	
	//handle program
	if (program) {
		printf("Programming flash ");
		fflush(stdout);
		dotCounter = 18;
		addr = FLASH_ADDR + offset;
		for (i=0 ; i<count ; i+=FLASH_PAGE_SIZE) { //iterate over pages
			for (j=0 ; j<FLASH_PAGE_SIZE ; j+=4) { //iterate over words
				data  =  buffer[i+j]   & 0xff;
				data |= (buffer[i+j+1] & 0xff) <<  8;
				data |= (buffer[i+j+2] & 0xff) << 16;
				data |= (buffer[i+j+3] & 0xff) << 24;
				writeWord(usb, addr+i+j, data);
			}
			page = (offset + i) / FLASH_PAGE_SIZE;
			if (page >= getFlashPageCount(dev)) failExit("not enough flash space\n", -5);
			writeWord(usb, EEFC_FCR_ADDR, EEFC_FCR_CMD_WRITE_PAGE | (page << 8));
			waitFlashReady(usb);
			printf(".");
			if (++dotCounter > 80) {
				printf("\n");
				dotCounter = 0;
			}
			fflush(stdout);
		}
		printf(" OK\n");
	}
	
	//handle writing RAM
	if (writeRam) {
		printf("Writing to RAM ");
		fflush(stdout);
		dotCounter = 19;
		addr = RAM_ADDR + offset;
		for (i=0 ; i<count ; i+=4) { //iterate over words
			data  =  buffer[i]   & 0xff;
			data |= (buffer[i+1] & 0xff) <<  8;
			data |= (buffer[i+2] & 0xff) << 16;
			data |= (buffer[i+3] & 0xff) << 24;
			writeWord(usb, addr+i, data);
			if (i%512 == 0) {
				printf(".");
				if (++dotCounter > 80) {
					printf("\n");
					dotCounter = 0;
				}
				fflush(stdout);
			}
		}
		printf(" OK\n");
	}
	
	//handle remap flash
	if (remap) {
		printf("Remapping Flash ...");
		waitFlashReady(usb);
		writeWord(usb, EEFC_FCR_ADDR, EEFC_FCR_CMD_SET_GPNVM | (1 << 8)); //GPNVM 1 is boot mode selection (set to 1 for flash)
		waitFlashReady(usb);
		printf(" OK\n");
	}
	
	//handle reset
	if (reset) {
		printf("Resetting ...");
		writeWordNoResponse(usb, RSTC_CR_ADDR, RSTC_CR_CMD_RST_ALL);
		printf(" OK\n");
	}
	
	
	//clean up
	if (dev != NULL) {
		(*(usb.interface))->USBInterfaceClose(usb.interface);
		(*(usb.interface))->Release(usb.interface);
	}
}



//-----------------------------------------------------------------------------------------------------------------

