/*
 * MacSamba
 * <http://ghostwood.org/software/macSamba.html>
 * Copyright 2006-15 Nick Howes <nick@ghostwood.org>
 * tabs every 4 spaces
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <IOKit/usb/IOUSBLib.h>

#include "devices.h"
#include "usb.h"
#include "util.h"

#define REG_CHIPID_CIDR	0x400E0940U	// Chip ID register
#define REG_CHIPID_EXID	0x400E0944U	// Chip ID register

#define EEFC_FSR_ADDR	0x400E0C08	// Flash Status Register
#define EEFC_FSR_ERR	0x0000000E	// FLERR, FLOCKE, FCMDE


#define SLEEP_TIME 10000
#define RETRIES 1000

int readWord(usb_t usb, unsigned int addr, unsigned int* word) {
	char buf[256], *p;
	int n, r;

	n = snprintf(buf, sizeof(buf), "w%X,#", addr);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf));
	p = buf;
	while (isspace(*p)) p++; //skip any whitespace
	r = sscanf(p, "0x%x", word);
	if (r == 1) {
		return 0;
	} else {
		return -1;
	}
}

int readByte(usb_t usb, unsigned int addr, unsigned int* word) {
	char buf[256], *p;
	int n, r;

	n = snprintf(buf, sizeof(buf), "o%X,#", addr);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf));
	p = buf;
	while (isspace(*p)) p++; //skip any whitespace
	r = sscanf(p, "0x%x", word);
	if (r == 1) {
		return 0;
	} else {
		return -1;
	}
}

int writeWord(usb_t usb, unsigned int addr, unsigned int word) {
	char buf[256];
	int n;

	//printf("W%X,%X#\n", addr, word);
	n = snprintf(buf, sizeof(buf), "W%X,%08X#", addr, word);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf)); //consume response
	return 0;
}

int writeWordNoResponse(usb_t usb, unsigned int addr, unsigned int word) {
	char buf[256];
	int n;

	//printf("W%X,%X#\n", addr, word);
	n = snprintf(buf, sizeof(buf), "W%X,%08X#", addr, word);
	writeUSB(usb, buf, n);
	return 0;
}

int writeBlock(usb_t usb, unsigned int addr, char* buffer, unsigned int size) {
	int i;
	unsigned int data;
	
	for (i=0 ; i<size ; i+=4) {
		data = (*buffer) & 0xff;
		++buffer;
		data |= ((*buffer) & 0xff) << 8;
		++buffer;
		data |= ((*buffer) & 0xff) << 16;
		++buffer;
		data |= ((*buffer) & 0xff) << 24;
		++buffer;
		writeWord(usb, addr + i, data);
	}

	return 0;
}

deviceInfo_t* getDeviceInfo(usb_t usb) {
	unsigned int data;
	unsigned int extData;

	readWord(usb, REG_CHIPID_CIDR, &data);
	readWord(usb, REG_CHIPID_EXID, &extData);
//	printf(" ... device ID is %08X %08X\n", data, extData);

	return getDeviceByID(data, extData);
}

void waitFlashReady(usb_t usb) {
	int retry = 0;
	unsigned int data;

	do {
		readWord(usb, EEFC_FSR_ADDR, &data);
		if (data & EEFC_FSR_ERR) {
			printf("Flash Error 0X%X\n", data);
		}
		usleep(SLEEP_TIME);
		if ((data & 1) == 1) return;
	} while (++retry < RETRIES);
	
	failExit("timeout waiting for flash to become ready", 0);
}



